import snippetfind
import stripogram
import re
from OFS.SimpleItem import SimpleItem
from AccessControl import ClassSecurityInfo, getSecurityManager
from Globals import InitializeClass
from Products.PageTemplates.PageTemplateFile import PageTemplateFile

from Products.Silva.helpers import add_and_edit
from Products.Silva import SilvaPermissions

class SearchService(SimpleItem):
    security = ClassSecurityInfo()

    meta_type = "Silva Search Service"
    
    manage_options = (
        {'label':'Edit', 'action':'manage_main'},
        ) + SimpleItem.manage_options
    
    def __init__(self, id, title):
        self.id = id
        self.title = title
        self._snippet_size = 40
        self._snippet_amount = 3
        self._start_tag = '<strong>'
        self._end_tag = '</strong>'

    manage_main = manage_editForm = PageTemplateFile(
        'www/searchServiceEdit', globals(),
        __name__='manage_editForm')
    
    security.declareProtected('View management screens', 'manage_edit')
    def manage_edit(self, snippet_size, snippet_amount, start_tag, end_tag):
        """Change the settings.
        """
        self._snippet_size = int(snippet_size)
        self._snippet_amount = int(snippet_amount)
        self._start_tag = start_tag
        self._end_tag = end_tag
        return self.manage_editForm(manage_tabs_message="Changed settings")

    security.declareProtected('View management screens', 'get_snippet_size')
    def get_snippet_size(self):
        '''snippet size'''
        return self._snippet_size
    
    security.declareProtected('View management screens', 'get_snippet_amount')
    def get_snippet_amount(self):
        '''snippet amount per object'''
        return self._snippet_amount

    security.declareProtected('View management screens', 'get_start_tag')
    def get_start_tag(self):
        '''start tag for query term'''
        return self._start_tag

    security.declareProtected('View management screens', 'get_end_tag')
    def get_end_tag(self):
        '''end tag for query term'''
        return self._end_tag
    
    security.declareProtected(SilvaPermissions.View,
                              'search')
    def search(self, fulltext, root=None, **extra):
        """Search in objects for fulltext.
        """
        # get rid of NOT as this leads to Zope parse error
        fulltext = fulltext.replace('NOT', '')
        q = {
            'fulltext': fulltext,
            'meta_type': ['Silva Document Version', 'Silva Ghost Version'],
            'version_status': ['public'],
            }
        if root is not None:
            q['path'] = '/'.join(root.getPhysicalPath())
        if extra is not None:
            q.update(extra)
        return self.service_catalog.searchResults(q)

    security.declareProtected(SilvaPermissions.View,
                              'getBatchResults')
    def getSearchBatch(self, batch, fulltext):
        """Get search batch with extra information.
        """
        result = []
        query = ZCTextIndexQuery(fulltext)
        security_manager = getSecurityManager()
        for brain in batch:
            object = brain.getObject()
            # we can't really show objects that cannot be found;
            # catalog problem
            if object is None:
                # XXX log something here?
                continue
            
            # return None for items which are not visible
            if not security_manager.checkPermission('View', object):
                result.append(None)
                continue
            
            text = cleanupHTML(object.aq_parent.view())
            snippet = snippetfind.highlightedSnippet(
                query, text,
                self._snippet_size,
                self._snippet_amount,
                self._start_tag,
                self._end_tag)
            # XXX hack, shouldn't be in generic product code but
            # somehow factored out
            container = object.get_container()
            if (container.id == 'related_content' or
                container.get_title() == 'related_content'):   
                # get parent container and its url
                object = container.aq_parent
                url = object.absolute_url()
            else:
                url = stripurl(object.absolute_url())
            title = object.get_title()
            if title.strip() == '':
                title = 'no title'
            item = SearchItem(
                title=title,
                snippet=snippet,
                url=url,
                object=object)
            result.append(item)
        return result
  
InitializeClass(SearchService)

class SearchItem:
    __allow_access_to_unprotected_subobjects__ = 1
    
    def __init__(self, **kw):
        self.__dict__.update(kw)
        

class ZCTextIndexQuery(snippetfind.Query):
    """A special query class that can deal with ZCTextIndex query patterns
    """
    def getPhraseInfo(self, phrase):
        phrase = phrase.replace('AND', '')
        phrase = phrase.replace('OR', '')
        phrase = phrase.replace('NOT', '')
        phrase = phrase.strip().lower()
        # find all phrases in double quotes
        fullphrases = re.findall(r'"([^"]+)"', phrase)
        if fullphrases:
            result = []
            for fullphrase in fullphrases:
                phrase = phrase.replace('"%s"' % fullphrase, '')
                result.append((fullphrase, 3))
            for subphrase in phrase.split():
                result.append((subphrase, 1))
            return result
        # no full phrases, so use traditional strategy
        return snippetfind.Query.getPhraseInfo(self, phrase)    
    
def stripurl(url):
    """Strip version number, index off end of URL.
    """
    i = url.rfind('/')
    if i == -1:
        return url
    url = url[:i]
    i = url.rfind('/index')
    if i == -1:
        return url
    return url[:i]

def cleanupHTML(text):
    text = stripogram.html2text(text)
    text = ' '.join(text.split())
    return text

manage_addSearchServiceForm = PageTemplateFile(
    "www/searchServiceAdd", globals(),
    __name__ = 'manage_addSearchServiceForm')

def manage_addSearchService(self, id, title, REQUEST=None):
    """Add the search service"""
    id = self._setObject(id, SearchService(id, title))
    add_and_edit(self, id, REQUEST)
    return ''
    
