import re
            
class Query:
    def __init__(self, phrase):
        self._phrase = phrase
        subphrase_to_id = self.subphrase_to_id = {}
        id_to_subphrase = self.id_to_subphrase = []
        id_to_score = self.id_to_score = []
        i = 0
        for subphrase, score in self.getPhraseInfo(phrase):
            subphrase_to_id[subphrase] = i
            id_to_subphrase.append(subphrase)
            id_to_score.append(score)
            i += 1
        self._index_to_id = {}

    def getSubphrases(self):
        return self.subphrase_to_id.keys()

    def getPhraseInfo(self, phrase):
        """Return a list of subphrase, score tuples.
        """
        phrase = phrase.strip().lower()
        if not phrase:
            return []
        result = [(phrase, 3)]
        subphrases = phrase.split()
        if len(subphrases) == 1:
            return result
        for subphrase in subphrases:
            result.append((subphrase, 1))
        return result

class TextMap:
    def __init__(self, query, text):
        self._text = text
        self._query = query
        
        index_to_score = self.index_to_score = {}
        index_to_id = self.index_to_id = {}
        
        q = createSearchRe(query.getSubphrases())
        # handle case sensitivity
        text = text.lower()
        result = {}
        i = 0
        subphrase_to_id = query.subphrase_to_id
        id_to_score = query.id_to_score
        while 1:
            m = q.search(text, i)
            if m is None:
                break
            index = m.start(m.lastindex)
            id = subphrase_to_id[m.group(m.lastindex)]
            index_to_id[index] = id
            index_to_score[index] = id_to_score[id]
            i = m.end(m.lastindex)

    def getSubphraseForIndex(self, index):
        return self._query.id_to_subphrase[self.index_to_id[index]]

        
def createSearchRe(phrases):
    """Create regular expression to search all phrases.
    """
    phrases.sort(lambda x, y: cmp(len(y), len(x)))
    phrases = [r'(?:\W|^)(%s)(?:\W|$)' % re.escape(s) for s in phrases]
    return re.compile('|'.join(phrases), re.IGNORECASE | re.MULTILINE)

class Partition:
    def __init__(self, text_map, score, pos, size, indexes):
        self._text_map = text_map
        self._score = score
        self._pos = pos
        self._size = size
        indexes.sort()
        self._indexes = indexes
        
    def getPartition(self, pos, size):
        score = 0
        indexes = []
        for i, value in self._text_map.index_to_score.items():
            oi = i
            i = i - pos
            if i < 0 or i >= self._size:
                continue
            indexes.append(oi)
            score += value
        if score == 0:
            return NullPartition(pos, size)
        else:
            return Partition(self._text_map, score, pos, size, indexes)
    
    def getPartitions(self, partitions_amount):
        partition_size = int(self._size / partitions_amount)
        partition_map = {}
        for i, value in self._text_map.index_to_score.items():
            oi = i
            i = i - self._pos
            if i < 0 or i >= self._size:
                continue
            partition_id = int(i / partition_size)
            score, indexes = partition_map.get(partition_id, (0, []))
            indexes.append(oi)
            score += value
            partition_map[partition_id] = score, indexes
        result = []
        for partition_id, (score, indexes) in partition_map.items():
            pos = self._pos + partition_id * partition_size
            partition = Partition(self._text_map, score,
                                  pos, partition_size, indexes)
            result.append(partition)
        return result

    def getTopScorePartitions(self, partitions_amount, amount,
                              minimum=1):
        partitions = self.getPartitions(partitions_amount)
        d = {}
        for partition in partitions:
            d.setdefault(partition.getScore(), []).append(
                partition)
        scores = d.keys()
        scores.sort()
        result = []
        while len(result) < amount and scores:
            result.extend(d[scores.pop()])
        if len(result) > amount:
            result = result[:amount]
        result = [partition for partition in result
                  if partition.getScore() >= minimum]
        result.sort(lambda x, y: cmp(x.getPosition(), y.getPosition()))
        return result

    def getTopScorePartitionsUseOffset(self, partitions_amount, amount,
                                       minimum=1, offset_steps=5):
        partition_size = self._size / partitions_amount
        partitions = self.getTopScorePartitions(
            partitions_amount, amount, minimum)
        for step in range(2, offset_steps):
            offset = partition_size / step
            offset_partition = self.getPartition(offset,
                                                 self._size - offset)
            partitions.extend(offset_partition.getTopScorePartitions(
                partitions_amount, amount, minimum))
        # sort resulting partitions by score
        partitions.sort(lambda x, y: cmp(y.getScore(), x.getScore()))
        #partitions = partitions[:amount]
        #partitions.sort(lambda x, y: cmp(x.getPosition(), y.getPosition()))
        return partitions

    def getTopScorePartitionsNoOverlap(self, partitions_amount, amount,
                                       minimum=1):
        partitions = self.getTopScorePartitionsUseOffset(partitions_amount,
                                                         amount, minimum)
        partitions = getTopPartitions(partitions, amount)
        partitions.sort(lambda x, y:
                        cmp(x.getPosition(), y.getPosition()))
        return partitions
    
    def getPosition(self):
        return self._pos
    
    def getSize(self):
        return self._size

    def getRange(self):
        return self._pos, self._pos + self._size

    def getIndexes(self):
        return self._indexes
    
    def getExtendedRange(self, word_size=20):
        """
        Add to start to get to word barrier (or start of subphrase). Also add
        to end to get to word barrier (or end of phrase).
        """
        start, end = self.getRange()
        # add to the end to get the last phrase if necessary
        last_index = self._indexes[-1]
        subphrase = self._text_map.getSubphraseForIndex(last_index)
        if last_index + len(subphrase) >= end:
            end = last_index + len(subphrase)
        # add to the end to get a word XXX could also contract..
        pattern = r'[^\s]{,%s}' % word_size
        q = re.compile(pattern, re.MULTILINE)
        m = q.match(self._text_map._text, end)
        end = m.end()
        # look for new start; we don't need to if we're at start of text itself
        if start != 0:
            # in the beginning, get to a word too, unless we cross first index
            first_index = self._indexes[0]
            pattern = r'[^\s]{,%s}\s?' % word_size
            q = re.compile(pattern, re.MULTILINE)
            m = q.match(self._text_map._text, start)
            found_start = m.end()
            if found_start > first_index:
                start = first_index
            else:
                start = found_start
        return start, end

    def getScore(self):
        return self._score

    def getText(self):
        start, end = self.getExtendedRange()
        return self._text_map._text[start:end]

    def __repr__(self):
        return '<snippetfind.Partition range %s,%s>' % self.getRange()
    
class NullPartition(Partition):
    def __init__(self, pos, size):
        self._pos = pos
        self._size = size

    def getPartitions(self, partitions_amount):
        result = []
        partition_size = int(self._size / partitions_amount)
        for partition_id in range(partitions_amount):
            pos = self._pos + partition_id * partition_size
            result.append(NullPartition(pos, partition_size))
        return result
            
    def getTopScorePartitions(self, partitions_amount, amount, minimum=1):
        return []
    
    def getScore(self):
        return 0

    def getIndexes(self):
        return []

    def getExtendedRange(self):
        return self.getRange()

class RootPartition(Partition):
    def __init__(self, query, text):
        text_map = TextMap(query, text)
        score = 0
        for value in text_map.index_to_score.values():
            score += value
        indexes = text_map.index_to_score.keys()
        Partition.__init__(self, text_map, score, 0, len(text), indexes)

    def getTop(self, partition_size, amount, minimum=1):
        partition_amount = self.getSize() / partition_size
        if partition_amount == 0:
            partition_amount = 1
        return self.getTopScorePartitionsNoOverlap(partition_amount,
                                                   amount, minimum)

def getTopPartitions(partitions, amount):
    result = []
    getTopPartitionsHelper(partitions, result, amount)
    return result

def getTopPartitionsHelper(partitions, result, amount):
    if not partitions:
        return
    if len(result) >= amount:
        return
    top, rest = getTopRest(partitions)
    result.append(top)
    start, end = top.getExtendedRange()
    partitions = []
    for partition in rest:
        p_start, p_end = partition.getExtendedRange()
        # no overlap overlap
        if ((start <= p_start < end) or
            (start <= p_end < end) or
            (p_start < start and p_end >= end)):
            continue
        partitions.append(partition)
    getTopPartitionsHelper(partitions, result, amount)

def getTopRest(partitions):
    # we need to sort on score, but secondary condition is lowest position
    partitions.sort(lambda x, y: cmp(
        (x.getScore(), y.getPosition()),
        (y.getScore(), x.getPosition())))
    return partitions[-1], partitions[:-1]        

def reconstructText(text, partitions):
    """From ordered partitions, reconstruct text.
    """
    result = []
    for partition in partitions:
        start, end = partition.getRange()
        result.append(text[start:end])
    return ''.join(result)

def highlight(text, phrases, starttag, endtag):
    """Highlight all items in phrases found in text using starttag, endtag.

    If no items can be found, don't highlight anything. Text should be
    appropriately HTML stripped, HTML escaped and whitespace
    normalized if necessary by caller.
    """
    # if there is nothing to highlight, do nothing
    if not phrases:
        return text
    q = createSearchRe(phrases)
    i = 0
    pairs = []
    while i < len(text):
        m = q.search(text, i)
        if m is None:
            break
        pairs.append((m.start(), m.end()))
        i = m.end() + 1
    if not pairs:
        return text
    result = []
    last_end = 0
    for start, end in pairs:
        result.append(text[last_end:start])
        result.append(starttag)
        result.append(text[start:end])
        result.append(endtag)
        last_end = end
    result.append(text[last_end:])
    return ''.join(result)

def highlightedSnippet(query, text, partition_size, amount, starttag, endtag):
    root = RootPartition(query, text)
    partitions = root.getTop(partition_size, amount)
    subphrases = query.getSubphrases()
    text_result = []
    for partition in partitions:
        text_result.append(partition.getText())
    text = ' ... '.join(text_result)
    return highlight(text, subphrases, starttag, endtag)
    
def main():
    f = open('test.txt', 'r')
    text = f.read()
    f.close()
    #phrase = 'lamp chandelier torch candle'
    phrase = 'test'
    root = RootPartition(Query(phrase), text)
    for partition in root.getTop(100, 3):
        print partition.getRange(), partition.getExtendedRange(), partition.getScore()
        start, end = partition.getExtendedRange()
        print text[start:end]
        print "=" * 40

if __name__ == '__main__':
    main()
    
